from datetime import timedelta

from torch._C._distributed_c10d import _DEFAULT_PG_TIMEOUT


__all__ = ["default_pg_timeout", "default_pg_nccl_timeout"]

# Default process group wide timeout, if applicable.
# This only applies to the non-nccl backends
# To make an attempt at backwards compatibility with THD, we use an
# extraordinarily high default timeout, given that THD did not have timeouts.
default_pg_timeout: timedelta = _DEFAULT_PG_TIMEOUT
# Separate timeout for PGNCCL mainly because it's always been that way in the C++ layer, but until recently
# there was one default that applied across all backends in the python layer.
# Later, we could consider merging them back together at the c++ layer if we can align on a same value.
# (only if TORCH_NCCL_BLOCKING_WAIT or TORCH_NCCL_ASYNC_ERROR_HANDLING is set to 1).

try:
    from torch._C._distributed_c10d import _DEFAULT_PG_NCCL_TIMEOUT

    default_pg_nccl_timeout: timedelta | None = _DEFAULT_PG_NCCL_TIMEOUT
except ImportError:
    # if C++ NCCL support is not compiled, we don't have access to the default nccl value.
    # if anyone is actually trying to use nccl in this state, it should error.
    default_pg_nccl_timeout = None



# mypy: allow-untyped-defs
import ast
import builtins
import dis
import enum
import inspect
import re
import typing
import warnings
from textwrap import dedent

import torch
from torch._C import (
    _GeneratorType,
    AnyType,
    AwaitType,
    BoolType,
    ComplexType,
    DeviceObjType,
    DictType,
    EnumType,
    FloatType,
    FutureType,
    InterfaceType,
    IntType,
    ListType,
    NoneType,
    NumberType,
    OptionalType,
    StreamObjType,
    StringType,
    TensorType,
    TupleType,
    UnionType,
)
from torch._jit_internal import (
    _Await,
    _qualified_name,
    Any,
    BroadcastingList1,
    BroadcastingList2,  # pyrefly: ignore [missing-module-attribute]
    BroadcastingList3,  # pyrefly: ignore [missing-module-attribute]
    Dict,
    Future,
    is_await,
    is_dict,
    is_future,
    is_ignored_fn,
    is_list,
    is_optional,
    is_tuple,
    is_union,
    List,
    Optional,
    Tuple,
    Union,
)
from torch._sources import get_source_lines_and_file

from ._state import _get_script_class


if torch.distributed.rpc.is_available():
    from torch._C import RRefType
    from torch._jit_internal import is_rref, RRef

from torch._ops import OpOverloadPacket


class Module:
    def __init__(self, name, members) -> None:
        self.name = name
        self.members = members

    def __getattr__(self, name):
        try:
            return self.members[name]
        except KeyError:
            raise RuntimeError(
                f"Module {self.name} has no member called {name}"
            ) from None


class EvalEnv:
    env = {
        "torch": Module("torch", {"Tensor": torch.Tensor}),
        "Tensor": torch.Tensor,
        "typing": Module("typing", {"Tuple": Tuple}),
        "Tuple": Tuple,
        "List": List,
        "Dict": Dict,
        "Optional": Optional,
        "Union": Union,
        "Future": Future,
        "Await": _Await,
    }

    def __init__(self, rcb) -> None:
        self.rcb = rcb
        if torch.distributed.rpc.is_available():
            # pyrefly: ignore [bad-typed-dict-key, unsupported-operation]
            self.env["RRef"] = RRef

    def __getitem__(self, name):
        if name in self.env:
            return self.env[name]
        if self.rcb is not None:
            return self.rcb(name)
        return getattr(builtins, name, None)


def get_signature(fn, rcb, loc, is_method):
    if isinstance(fn, OpOverloadPacket):
        signature = try_real_annotations(fn.op, loc)
    else:
        signature = try_real_annotations(fn, loc)
    if signature is not None and is_method:
        # If this is a method, then the signature will include a type for
        # `self`, but type comments do not contain a `self`. So strip it
        # away here so everything is consistent (`inspect.ismethod` does
        # not work here since `fn` is unbound at this point)
        param_types, return_type = signature
        param_types = param_types[1:]
        signature = (param_types, return_type)

    if signature is None:
        type_line, source = None, None
        try:
            source = dedent("".join(get_source_lines_and_file(fn)[0]))
            type_line = get_type_line(source)
        except TypeError:
            pass
        # This might happen both because we failed to get the source of fn, or
        # because it didn't have any annotations.
        if type_line is not None:
            signature = parse_type_line(type_line, rcb, loc)

    return signature


def is_function_or_method(the_callable):
    # A stricter version of `inspect.isroutine` that does not pass for built-in
    # functions
    return inspect.isfunction(the_callable) or inspect.ismethod(the_callable)


def is_vararg(the_callable):
    if not is_function_or_method(the_callable) and callable(the_callable):  # noqa: B004
        # If `the_callable` is a class, de-sugar the call so we can still get
        # the signature
        the_callable = the_callable.__call__

    if is_function_or_method(the_callable):
        return inspect.getfullargspec(the_callable).varargs is not None
    else:
        return False


def get_param_names(fn, n_args):
    if isinstance(fn, OpOverloadPacket):
        fn = fn.op

    if (
        not is_function_or_method(fn)
        and callable(fn)
        and is_function_or_method(fn.__call__)
    ):  # noqa: B004
        # De-sugar calls to classes
        fn = fn.__call__

    if is_function_or_method(fn):
        if is_ignored_fn(fn):
            fn = inspect.unwrap(fn)
        return inspect.getfullargspec(fn).args
    else:
        # The `fn` was not a method or function (maybe a class with a __call__
        # method, so use a default param name list)
        return [str(i) for i in range(n_args)]


def check_fn(fn, loc) -> None:
    # Make sure the function definition is not a class instantiation
    try:
        source = dedent("".join(get_source_lines_and_file(fn)[0]))
    except (OSError, TypeError):
        return
    if source is None:
        return

    py_ast = ast.parse(source)
    if len(py_ast.body) == 1 and isinstance(py_ast.body[0], ast.ClassDef):
        raise torch.jit.frontend.FrontendError(
            loc,
            f"Cannot instantiate class '{py_ast.body[0].name}' in a script function",
        )
    if len(py_ast.body) != 1 or not isinstance(py_ast.body[0], ast.FunctionDef):
        raise torch.jit.frontend.FrontendError(
            loc, "Expected a single top-level function"
        )


def _eval_no_call(stmt, glob, loc):
    """Evaluate statement as long as it does not contain any method/function calls."""
    bytecode = compile(stmt, "", mode="eval")
    for insn in dis.get_instructions(bytecode):
        if "CALL" in insn.opname:
            raise RuntimeError(
                f"Type annotation should not contain calls, but '{stmt}' does"
            )
    return eval(bytecode, glob, loc)  # type: ignore[arg-type] # noqa: P204


def parse_type_line(type_line, rcb, loc):
    """Parse a type annotation specified as a comment.

    Example inputs:
        # type: (Tensor, torch.Tensor) -> Tuple[Tensor]
        # type: (Tensor, Tuple[Tensor, Tensor]) -> Tensor
    """
    arg_ann_str, ret_ann_str = split_type_line(type_line)

    try:
        arg_ann = _eval_no_call(arg_ann_str, {}, EvalEnv(rcb))
    except (NameError, SyntaxError) as e:
        raise RuntimeError(
            "Failed to parse the argument list of a type annotation"
        ) from e

    if not isinstance(arg_ann, tuple):
        arg_ann = (arg_ann,)

    try:
        ret_ann = _eval_no_call(ret_ann_str, {}, EvalEnv(rcb))
    except (NameError, SyntaxError) as e:
        raise RuntimeError(
            "Failed to parse the return type of a type annotation"
        ) from e

    arg_types = [ann_to_type(ann, loc) for ann in arg_ann]
    return arg_types, ann_to_type(ret_ann, loc)


def get_type_line(source):
    """Try to find the line containing a comment with the type annotation."""
    type_comment = "# type:"

    lines = source.split("\n")
    lines = list(enumerate(lines))
    type_lines = list(filter(lambda line: type_comment in line[1], lines))
    # `type: ignore` comments may be needed in JIT'ed functions for mypy, due
    # to the hack in torch/_VF.py.

    # An ignore type comment can be of following format:
    #   1) type: ignore
    #   2) type: ignore[rule-code]
    # This ignore statement must be at the end of the line

    # adding an extra backslash before the space, to avoid triggering
    # one of the checks in .github/workflows/lint.yml
    type_pattern = re.compile("# type:\\ ignore(\\[[a-zA-Z-]+\\])?$")
    type_lines = list(filter(lambda line: not type_pattern.search(line[1]), type_lines))

    if len(type_lines) == 0:
        # Catch common typo patterns like extra spaces, typo in 'ignore', etc.
        wrong_type_pattern = re.compile("#[\t ]*type[\t ]*(?!: ignore(\\[.*\\])?$):")
        wrong_type_lines = list(
            filter(lambda line: wrong_type_pattern.search(line[1]), lines)
        )
        if len(wrong_type_lines) > 0:
            raise RuntimeError(
                "The annotation prefix in line "
                + str(wrong_type_lines[0][0])
                + " is probably invalid.\nIt must be '# type:'"
                + "\nSee PEP 484 (https://www.python.org/dev/peps/pep-0484/#suggested-syntax-for-python-2-7-and-straddling-code)"  # noqa: B950
                + "\nfor examples"
            )
        return None
    elif len(type_lines) == 1:
        # Only 1 type line, quit now
        return type_lines[0][1].strip()

    # Parse split up argument types according to PEP 484
    # https://www.python.org/dev/peps/pep-0484/#suggested-syntax-for-python-2-7-and-straddling-code
    return_line = None
    parameter_type_lines = []
    for line_num, line in type_lines:
        if "# type: (...) -> " in line:
            return_line = (line_num, line)
            break
        elif type_comment in line:
            parameter_type_lines.append(line)
    if return_line is None:
        raise RuntimeError(
            "Return type line '# type: (...) -> ...' not found on multiline "
            "type annotation\nfor type lines:\n"
            + "\n".join([line[1] for line in type_lines])
            + "\n(See PEP 484 https://www.python.org/dev/peps/pep-0484/#suggested-syntax-for-python-2-7-and-straddling-code)"
        )

    def get_parameter_type(line):
        item_type = line[line.find(type_comment) + len(type_comment) :]
        return item_type.strip()

    types = map(get_parameter_type, parameter_type_lines)
    parameter_types = ", ".join(types)

    return return_line[1].replace("...", parameter_types)


def split_type_line(type_line):
    """Split the comment with the type annotation into parts for argument and return types.

    For example, for an input of:
        # type: (Tensor, torch.Tensor) -> Tuple[Tensor, Tensor]

    This function will return:
        ("(Tensor, torch.Tensor)", "Tuple[Tensor, Tensor]")

    """
    start_offset = len("# type:")
    try:
        arrow_pos = type_line.index("->")
    except ValueError:
        raise RuntimeError(
            "Syntax error in type annotation (couldn't find `->`)"
        ) from None
    return type_line[start_offset:arrow_pos].strip(), type_line[arrow_pos + 2 :].strip()


def try_real_annotations(fn, loc):
    """Try to use the Py3.5+ annotation syntax to get the type."""
    try:
        # Note: anything annotated as `Optional[T]` will automatically
        # be returned as `Union[T, None]` per
        # https://github.com/python/cpython/blob/main/Lib/typing.py#L732
        sig = inspect.signature(fn)
    except ValueError:
        return None

    all_annots = [sig.return_annotation] + [
        p.annotation for p in sig.parameters.values()
    ]
    if all(ann is sig.empty for ann in all_annots):
        return None

    arg_types = [ann_to_type(p.annotation, loc) for p in sig.parameters.values()]
    return_type = ann_to_type(sig.return_annotation, loc)
    return arg_types, return_type


# Finds common type for enum values belonging to an Enum class. If not all
# values have the same type, AnyType is returned.
def get_enum_value_type(e: type[enum.Enum], loc):
    enum_values: List[enum.Enum] = list(e)
    if not enum_values:
        raise ValueError(f"No enum values defined for: '{e.__class__}'")

    types = {type(v.value) for v in enum_values}
    ir_types = [try_ann_to_type(t, loc) for t in types]

    # If Enum values are of different types, an exception will be raised here.
    # Even though Python supports this case, we chose to not implement it to
    # avoid overcomplicate logic here for a rare use case. Please report a
    # feature request if you find it necessary.
    res = torch._C.unify_type_list(ir_types)
    if not res:
        return AnyType.get()
    return res


def is_tensor(ann) -> bool:
    if issubclass(ann, torch.Tensor):
        return True

    if issubclass(
        ann,
        (
            torch.LongTensor,
            torch.DoubleTensor,
            torch.FloatTensor,
            torch.IntTensor,
            torch.ShortTensor,
            torch.HalfTensor,
            torch.CharTensor,
            torch.ByteTensor,
            torch.BoolTensor,
        ),
    ):
        warnings.warn(
            "TorchScript will treat type annotations of Tensor "
            "dtype-specific subtypes as if they are normal Tensors. "
            "dtype constraints are not enforced in compilation either.",
            stacklevel=2,
        )
        return True

    return False


def _fake_rcb(inp) -> None:
    return None


def try_ann_to_type(ann, loc, rcb=None):
    ann_args = typing.get_args(ann)  # always returns a tuple!

    if ann is inspect.Signature.empty:
        return TensorType.getInferred()
    if ann is None:
        return NoneType.get()
    if inspect.isclass(ann) and is_tensor(ann):
        return TensorType.get()
    if is_tuple(ann):
        # Special case for the empty Tuple type annotation `Tuple[()]`
        if len(ann_args) == 1 and ann_args[0] == ():
            return TupleType([])
        return TupleType([try_ann_to_type(a, loc) for a in ann_args])
    if is_list(ann):
        elem_type = try_ann_to_type(ann_args[0], loc)
        if elem_type:
            return ListType(elem_type)
    if is_dict(ann):
        key = try_ann_to_type(ann_args[0], loc)
        value = try_ann_to_type(ann_args[1], loc)
        # Raise error if key or value is None
        if key is None:
            raise ValueError(
                f"Unknown type annotation: '{ann_args[0]}' at {loc.highlight()}"
            )
        if value is None:
            raise ValueError(
                f"Unknown type annotation: '{ann_args[1]}' at {loc.highlight()}"
            )
        return DictType(key, value)
    if is_optional(ann):
        if issubclass(ann_args[1], type(None)):
            contained = ann_args[0]
        else:
            contained = ann_args[1]
        valid_type = try_ann_to_type(contained, loc)
        if not valid_type:
            raise AssertionError(
                f"Unsupported annotation {repr(ann)} could not be resolved because "
                f"{repr(contained)} could not be resolved. At\n{repr(loc)}"
            )
        return OptionalType(valid_type)
    if is_union(ann):
        # TODO: this is hack to recognize NumberType
        if set(ann_args) == {int, float, complex}:
            return NumberType.get()
        inner: List = []
        # We need these extra checks because both `None` and invalid
        # values will return `None`
        # TODO: Determine if the other cases need to be fixed as well
        for a in typing.get_args(ann):
            if a is None:
                inner.append(NoneType.get())
            maybe_type = try_ann_to_type(a, loc)
            if not maybe_type:
                raise AssertionError(
                    f"Unsupported annotation {repr(ann)} could not be resolved because "
                    f"{repr(a)} could not be resolved. At\n{repr(loc)}"
                )
            inner.append(maybe_type)
        return UnionType(inner)  # type: ignore[arg-type]
    if torch.distributed.rpc.is_available() and is_rref(ann):
        return RRefType(try_ann_to_type(ann_args[0], loc))
    if is_future(ann):
        return FutureType(try_ann_to_type(ann_args[0], loc))
    if is_await(ann):
        elementType = try_ann_to_type(ann_args[0], loc) if ann_args else AnyType.get()
        return AwaitType(elementType)
    if ann is float:
        return FloatType.get()
    if ann is complex:
        return ComplexType.get()
    if ann is int or ann is torch.SymInt:
        return IntType.get()
    if ann is str:
        return StringType.get()
    if ann is bool:
        return BoolType.get()
    if ann is Any:
        return AnyType.get()
    if ann is type(None):
        return NoneType.get()
    if inspect.isclass(ann) and hasattr(ann, "__torch_script_interface__"):
        return InterfaceType(ann.__torch_script_interface__)
    if ann is torch.device:
        return DeviceObjType.get()
    if ann is torch.Generator:
        return _GeneratorType.get()
    if ann is torch.Stream:
        return StreamObjType.get()
    if ann is torch.dtype:
        return IntType.get()  # dtype not yet bound in as its own type
    if ann is torch.qscheme:
        return IntType.get()  # qscheme not yet bound in as its own type

    if inspect.isclass(ann) and issubclass(ann, enum.Enum):
        if _get_script_class(ann) is None:
            scripted_class = torch.jit._script._recursive_compile_class(ann, loc)
            name = scripted_class.qualified_name()
        else:
            name = _qualified_name(ann)
        return EnumType(name, get_enum_value_type(ann, loc), list(ann))
    if inspect.isclass(ann):
        maybe_script_class = _get_script_class(ann)
        if maybe_script_class is not None:
            return maybe_script_class
        if torch._jit_internal.can_compile_class(ann):
            return torch.jit._script._recursive_compile_class(ann, loc)

    # Maybe resolve a NamedTuple to a Tuple Type
    if rcb is None:
        rcb = _fake_rcb
    # pyrefly: ignore [bad-argument-type]
    return torch._C._resolve_type_from_object(ann, loc, rcb)


def ann_to_type(ann, loc, rcb=None):
    the_type = try_ann_to_type(ann, loc, rcb)
    if the_type is not None:
        return the_type
    raise ValueError(f"Unknown type annotation: '{ann}' at {loc.highlight()}")


__all__ = [
    "Any",
    "List",
    "BroadcastingList1",
    "BroadcastingList2",
    "BroadcastingList3",
    "Tuple",
    "is_tuple",
    "is_list",
    "Dict",
    "is_dict",
    "is_optional",
    "is_union",
    "TensorType",
    "TupleType",
    "FloatType",
    "ComplexType",
    "IntType",
    "ListType",
    "StringType",
    "DictType",
    "AnyType",
    "Module",
    # TODO: Consider not exporting these during wildcard import (reserve
    # that for the types; for idiomatic typing code.)
    "get_signature",
    "check_fn",
    "get_param_names",
    "parse_type_line",
    "get_type_line",
    "split_type_line",
    "try_real_annotations",
    "try_ann_to_type",
    "ann_to_type",
]



# Owner(s): ["oncall: distributed checkpointing"]

import os
import shutil
import tempfile
from concurrent.futures import Future
from unittest.mock import Mock

import torch
from torch.distributed.checkpoint._experimental.checkpoint_process import (
    CheckpointProcess,
    CheckpointProcessConfig,
)
from torch.distributed.checkpoint._experimental.checkpoint_reader import (
    CheckpointReader,
)
from torch.distributed.checkpoint._experimental.checkpoint_writer import (
    CheckpointWriter,
    CheckpointWriterConfig,
)
from torch.distributed.checkpoint._experimental.checkpointer import (
    AsyncCheckpointer,
    Checkpointer,
    SyncCheckpointer,
)
from torch.distributed.checkpoint._experimental.staging import (
    CheckpointStagerConfig,
    DefaultStager,
)
from torch.distributed.checkpoint._experimental.types import RankInfo
from torch.testing._internal.common_utils import run_tests, TestCase


def subprocess_init_fn(name: str, parent_pid: int) -> None:
    """Initialize the subprocess for async checkpointer tests."""
    if name != "test-async-checkpointer":
        raise AssertionError(f"Unexpected subprocess name: {name}")
    if os.getpid() == parent_pid:
        raise AssertionError("This was supposed to run in a different process")
    if os.getppid() != parent_pid:
        raise AssertionError("This was supposed to run as a child to main process")


def ckpt_writer_init_fn(**kwargs) -> CheckpointWriter:
    """Initialize a CheckpointWriter in the subprocess."""
    return CheckpointWriter(
        config=kwargs.get("config"),
        rank_info=kwargs.get("rank_info"),
    )


class TestCheckpointer(TestCase):
    """Parameterized tests that work with both sync and async checkpointers."""

    def setUp(self):
        super().setUp()
        # Create a temporary directory for checkpoints
        self.temp_dir = tempfile.mkdtemp()

        # Create real objects for testing
        self.rank_info = RankInfo(
            global_world_size=1,
            global_rank=0,
        )
        self.writer_config = CheckpointWriterConfig()

        # Create reader for testing
        self.reader = CheckpointReader(
            rank_info=self.rank_info,
        )

        # Create test state dictionary
        self.state_dict = {
            "model": torch.nn.Linear(10, 5).state_dict(),
            "optimizer": {"param_groups": [{"lr": 0.01}]},
            "epoch": 5,
            "step": 1000,
        }

    def tearDown(self):
        # Clean up the temporary directory
        shutil.rmtree(self.temp_dir)

    def _create_sync_checkpointer(self) -> SyncCheckpointer:
        """Create a synchronous checkpointer."""
        writer = CheckpointWriter(
            config=self.writer_config,
            rank_info=self.rank_info,
        )
        return SyncCheckpointer(writer, self.reader)

    def _create_async_checkpointer(self) -> AsyncCheckpointer:
        """Create an asynchronous checkpointer."""
        # Create staging config for async operations
        # Use conservative settings to avoid CUDA issues in test environment
        stager_config = CheckpointStagerConfig(
            use_async_staging=True,
            use_pinned_memory=False,  # Disable to avoid CUDA memory issues
            use_shared_memory=True,
            use_non_blocking_copy=False,  # Disable to avoid CUDA issues
        )

        # Create process config
        process_config = CheckpointProcessConfig(
            subprocess_init_timeout_secs=30,
            subprocess_shutdown_timeout_secs=60,
        )

        # Create stager
        checkpoint_stager = DefaultStager(stager_config)

        # Create checkpoint process
        checkpoint_process = CheckpointProcess(
            rank_info=self.rank_info,
            config=process_config,
            subprocess_init_fn=subprocess_init_fn,
            subprocess_init_args=(
                "test-async-checkpointer",
                os.getpid(),
            ),
            checkpoint_writer_init_fn=ckpt_writer_init_fn,
            checkpoint_writer_init_args={
                "config": self.writer_config,
                "rank_info": self.rank_info,
            },
        )

        # Wait for process initialization
        checkpoint_process.process_creation_future.result()

        return AsyncCheckpointer(
            checkpoint_stager=checkpoint_stager,
            checkpoint_process=checkpoint_process,
            reader=self.reader,
        )

    def _get_checkpointers(self):
        """Get both sync and async checkpointers for parameterized testing."""
        return [
            ("sync", self._create_sync_checkpointer()),
            ("async", self._create_async_checkpointer()),
        ]

    def _save_checkpoint(self, checkpointer: Checkpointer, path, state_dict, **kwargs):
        """Save checkpoint and handle both sync/async return values."""
        result = checkpointer.save(path, state_dict, **kwargs)
        return (None, None) if result is None else result

    def _wait_for_save(self, stage_future, write_future):
        """Wait for save operation to complete."""
        if write_future is not None:
            write_future.result()
        if stage_future is not None:
            stage_future.result()

    def test_save_and_load_basic(self):
        """Test basic save and load functionality for both sync and async."""
        for checkpointer_type, checkpointer in self._get_checkpointers():
            with self.subTest(checkpointer_type=checkpointer_type):
                try:
                    checkpoint_path = os.path.join(
                        self.temp_dir, f"checkpoint_{checkpointer_type}"
                    )

                    # Save the checkpoint
                    stage_future, write_future = self._save_checkpoint(
                        checkpointer, checkpoint_path, self.state_dict
                    )
                    self._wait_for_save(stage_future, write_future)

                    # Verify that the checkpoint file exists
                    checkpoint_file = os.path.join(
                        checkpoint_path, f"checkpoint_{self.rank_info.global_rank}.pt"
                    )
                    self.assertTrue(os.path.exists(checkpoint_file))

                    # Load the checkpoint using the checkpointer
                    loaded_state_dict = checkpointer.load(checkpoint_path)

                    # Verify the loaded state dictionary
                    self.assertIn("model", loaded_state_dict)
                    self.assertIn("optimizer", loaded_state_dict)
                    self.assertEqual(loaded_state_dict["epoch"], 5)
                    self.assertEqual(loaded_state_dict["step"], 1000)

                finally:
                    checkpointer.close()

    def test_load_with_map_location(self):
        """Test loading with map_location for both sync and async."""
        for checkpointer_type, checkpointer in self._get_checkpointers():
            with self.subTest(checkpointer_type=checkpointer_type):
                try:
                    checkpoint_path = os.path.join(
                        self.temp_dir, f"checkpoint_map_{checkpointer_type}"
                    )

                    # Save the checkpoint
                    stage_future, write_future = self._save_checkpoint(
                        checkpointer, checkpoint_path, self.state_dict
                    )
                    self._wait_for_save(stage_future, write_future)

                    # Load with map_location
                    loaded_state_dict = checkpointer.load(
                        checkpoint_path, default_map_location="cpu"
                    )

                    # Verify the loaded state dictionary
                    self.assertIn("model", loaded_state_dict)
                    self.assertEqual(loaded_state_dict["epoch"], 5)

                finally:
                    checkpointer.close()

    def test_partial_load(self):
        """Test partial loading for both sync and async."""
        for checkpointer_type, checkpointer in self._get_checkpointers():
            with self.subTest(checkpointer_type=checkpointer_type):
                try:
                    checkpoint_path = os.path.join(
                        self.temp_dir, f"checkpoint_partial_{checkpointer_type}"
                    )

                    # Save the full checkpoint
                    stage_future, write_future = self._save_checkpoint(
                        checkpointer, checkpoint_path, self.state_dict
                    )
                    self._wait_for_save(stage_future, write_future)

                    # Create a partial state dictionary
                    partial_state_dict = {
                        "model": torch.nn.Linear(10, 5).state_dict(),
                        "epoch": None,
                    }

                    # Load only the keys in partial_state_dict
                    loaded_state_dict = checkpointer.load(
                        checkpoint_path, state_dict=partial_state_dict
                    )

                    # Verify partial loading worked
                    self.assertIn("model", loaded_state_dict)
                    self.assertIn("epoch", loaded_state_dict)
                    self.assertEqual(loaded_state_dict["epoch"], 5)
                    self.assertNotIn("step", loaded_state_dict)
                    self.assertNotIn("optimizer", loaded_state_dict)

                finally:
                    checkpointer.close()

    def test_load_strict_mode(self):
        """Test strict mode loading for both sync and async."""
        for checkpointer_type, checkpointer in self._get_checkpointers():
            with self.subTest(checkpointer_type=checkpointer_type):
                try:
                    checkpoint_path = os.path.join(
                        self.temp_dir, f"checkpoint_strict_{checkpointer_type}"
                    )

                    # Save a checkpoint with limited keys
                    limited_state_dict = {"model": torch.nn.Linear(10, 5).state_dict()}
                    stage_future, write_future = self._save_checkpoint(
                        checkpointer, checkpoint_path, limited_state_dict
                    )
                    self._wait_for_save(stage_future, write_future)

                    # Try to load with more keys than exist in checkpoint
                    partial_state_dict = {
                        "model": torch.nn.Linear(10, 5).state_dict(),
                        "missing_key": None,
                    }

                    # Should raise error in strict mode
                    with self.assertRaises(RuntimeError) as cm:
                        checkpointer.load(
                            checkpoint_path, state_dict=partial_state_dict, strict=True
                        )

                    self.assertIn("missing keys", str(cm.exception))

                    # Should work without strict mode
                    loaded_state_dict = checkpointer.load(
                        checkpoint_path, state_dict=partial_state_dict, strict=False
                    )
                    self.assertIn("model", loaded_state_dict)

                finally:
                    checkpointer.close()

    def test_save_with_kwargs(self):
        """Test save with additional kwargs for both sync and async."""
        for checkpointer_type, checkpointer in self._get_checkpointers():
            with self.subTest(checkpointer_type=checkpointer_type):
                try:
                    checkpoint_path = os.path.join(
                        self.temp_dir, f"checkpoint_kwargs_{checkpointer_type}"
                    )

                    # For sync checkpointer, we can pass arbitrary kwargs to the writer
                    # For async checkpointer, we test without kwargs to avoid conflicts
                    if checkpointer_type == "sync":
                        # Sync checkpointer passes kwargs directly to writer, so arbitrary kwargs are OK
                        stage_future, write_future = self._save_checkpoint(
                            checkpointer,
                            checkpoint_path,
                            self.state_dict,
                            custom_arg="test_value",
                            another_arg=42,
                        )
                    else:
                        # Async checkpointer has complex kwargs handling between stager and writer
                        # Just test basic save without kwargs to avoid conflicts
                        stage_future, write_future = self._save_checkpoint(
                            checkpointer,
                            checkpoint_path,
                            self.state_dict,
                        )

                    self._wait_for_save(stage_future, write_future)

                    # Verify checkpoint was created
                    checkpoint_file = os.path.join(
                        checkpoint_path, f"checkpoint_{self.rank_info.global_rank}.pt"
                    )
                    self.assertTrue(os.path.exists(checkpoint_file))

                finally:
                    checkpointer.close()

    def test_nested_dict_partial_load(self):
        """Test loading nested dictionaries partially for both sync and async."""
        for checkpointer_type, checkpointer in self._get_checkpointers():
            with self.subTest(checkpointer_type=checkpointer_type):
                try:
                    # Create a checkpoint with nested dictionaries
                    nested_state_dict = {
                        "model": {
                            "layer1": {
                                "weight": torch.randn(5, 10),
                                "bias": torch.randn(5),
                            },
                            "layer2": {
                                "weight": torch.randn(2, 5),
                                "bias": torch.randn(2),
                            },
                        },
                        "metadata": {"epoch": 10, "step": 2000},
                    }

                    checkpoint_path = os.path.join(
                        self.temp_dir, f"checkpoint_nested_{checkpointer_type}"
                    )

                    # Save the nested state dict
                    stage_future, write_future = self._save_checkpoint(
                        checkpointer, checkpoint_path, nested_state_dict
                    )
                    self._wait_for_save(stage_future, write_future)

                    # Create a partial state dictionary with nested structure
                    partial_state_dict = {
                        "model": {
                            "layer1": {"weight": None},  # Only request layer1.weight
                        },
                        "metadata": {"epoch": None},  # Only request metadata.epoch
                    }

                    # Load only the keys in partial_state_dict
                    loaded_state_dict = checkpointer.load(
                        checkpoint_path, state_dict=partial_state_dict
                    )

                    # Verify that the nested keys were correctly loaded
                    self.assertIn("model", loaded_state_dict)
                    self.assertIn("layer1", loaded_state_dict["model"])
                    self.assertIn("weight", loaded_state_dict["model"]["layer1"])
                    self.assertIn("metadata", loaded_state_dict)
                    self.assertIn("epoch", loaded_state_dict["metadata"])

                    # Verify values were loaded correctly
                    self.assertTrue(
                        torch.allclose(
                            loaded_state_dict["model"]["layer1"]["weight"],
                            nested_state_dict["model"]["layer1"]["weight"],
                        )
                    )
                    self.assertEqual(loaded_state_dict["metadata"]["epoch"], 10)

                    # Verify that keys not in the partial_state_dict are not loaded
                    self.assertNotIn("layer2", loaded_state_dict["model"])
                    self.assertNotIn("step", loaded_state_dict["metadata"])

                finally:
                    checkpointer.close()


class TestAsyncCheckpointerSpecific(TestCase):
    """Tests specific to AsyncCheckpointer functionality."""

    def setUp(self):
        super().setUp()
        # Create a temporary directory for checkpoints
        self.temp_dir = tempfile.mkdtemp()

        # Create real objects for testing
        self.rank_info = RankInfo(
            global_world_size=1,
            global_rank=0,
        )
        self.writer_config = CheckpointWriterConfig()

        # Create reader for testing
        self.reader = CheckpointReader(
            rank_info=self.rank_info,
        )

        # Create test state dictionary
        self.state_dict = {
            "model": torch.nn.Linear(10, 5).state_dict(),
            "optimizer": {"param_groups": [{"lr": 0.01}]},
            "epoch": 5,
            "step": 1000,
        }

    def tearDown(self):
        # Clean up the temporary directory
        shutil.rmtree(self.temp_dir)

    def _create_async_checkpointer(self) -> AsyncCheckpointer:
        """Helper method to create AsyncCheckpointer with real components."""
        # Create staging config for async operations
        # Use conservative settings to avoid CUDA issues in test environment
        stager_config = CheckpointStagerConfig(
            use_async_staging=True,
            use_pinned_memory=False,  # Disable to avoid CUDA memory issues
            use_shared_memory=True,
            use_non_blocking_copy=False,  # Disable to avoid CUDA issues
        )

        # Create process config
        process_config = CheckpointProcessConfig(
            subprocess_init_timeout_secs=30,
            subprocess_shutdown_timeout_secs=60,
        )

        # Create stager
        checkpoint_stager = DefaultStager(stager_config)

        # Create checkpoint process
        checkpoint_process = CheckpointProcess(
            rank_info=self.rank_info,
            config=process_config,
            subprocess_init_fn=subprocess_init_fn,
            subprocess_init_args=(
                "test-async-checkpointer",
                os.getpid(),
            ),
            checkpoint_writer_init_fn=ckpt_writer_init_fn,
            checkpoint_writer_init_args={
                "config": self.writer_config,
                "rank_info": self.rank_info,
            },
        )

        # Wait for process initialization
        checkpoint_process.process_creation_future.result()

        return AsyncCheckpointer(
            checkpoint_stager=checkpoint_stager,
            checkpoint_process=checkpoint_process,
            reader=self.reader,
        )

    def test_async_returns_futures(self):
        """Test that async save returns futures."""
        checkpointer = self._create_async_checkpointer()
        checkpoint_path = os.path.join(self.temp_dir, "checkpoint_futures")

        try:
            # Save the checkpoint asynchronously
            result = checkpointer.save(checkpoint_path, self.state_dict)

            # Verify that futures are returned
            self.assertIsInstance(result, tuple)
            self.assertEqual(len(result), 2)
            stage_future, write_future = result
            self.assertIsInstance(stage_future, Future)
            self.assertIsInstance(write_future, Future)

            # Wait for completion
            stage_future.result()
            write_future.result()

        finally:
            checkpointer.close()

    def test_async_sequential_saves_wait(self):
        """Test that sequential async saves wait for previous operations."""
        checkpointer = self._create_async_checkpointer()

        try:
            # First save
            checkpoint_path1 = os.path.join(self.temp_dir, "checkpoint_seq_1")
            stage_future1, write_future1 = checkpointer.save(
                checkpoint_path1, self.state_dict
            )

            # Second save (should wait for first to complete)
            checkpoint_path2 = os.path.join(self.temp_dir, "checkpoint_seq_2")
            modified_state_dict = self.state_dict.copy()
            modified_state_dict["epoch"] = 10
            stage_future2, write_future2 = checkpointer.save(
                checkpoint_path2, modified_state_dict
            )

            # Wait for both to complete
            write_future1.result()
            write_future2.result()

            # Verify both checkpoints were created with correct content
            checkpoint_file1 = os.path.join(
                checkpoint_path1, f"checkpoint_{self.rank_info.global_rank}.pt"
            )
            checkpoint_file2 = os.path.join(
                checkpoint_path2, f"checkpoint_{self.rank_info.global_rank}.pt"
            )

            self.assertTrue(os.path.exists(checkpoint_file1))
            self.assertTrue(os.path.exists(checkpoint_file2))

            loaded1 = torch.load(checkpoint_file1)
            loaded2 = torch.load(checkpoint_file2)

            self.assertEqual(loaded1["epoch"], 5)
            self.assertEqual(loaded2["epoch"], 10)

        finally:
            checkpointer.close()

    def test_async_multiple_saves_ordering(self):
        """Test that multiple async saves maintain proper ordering."""
        checkpointer = self._create_async_checkpointer()

        try:
            # Create multiple state dicts
            state_dicts = [
                {"epoch": 1, "model": torch.nn.Linear(5, 3).state_dict()},
                {"epoch": 2, "model": torch.nn.Linear(5, 3).state_dict()},
                {"epoch": 3, "model": torch.nn.Linear(5, 3).state_dict()},
            ]

            # Save multiple checkpoints
            futures = []
            checkpoint_paths = []
            for i, state_dict in enumerate(state_dicts, 1):
                checkpoint_path = os.path.join(self.temp_dir, f"multi_{i}")
                checkpoint_paths.append(checkpoint_path)
                stage_future, write_future = checkpointer.save(
                    checkpoint_path, state_dict
                )
                futures.append((stage_future, write_future))

            # Wait for all to complete
            for stage_future, write_future in futures:
                stage_future.result()
                write_future.result()

            # Verify all checkpoints exist and have correct content
            for i, checkpoint_path in enumerate(checkpoint_paths, 1):
                checkpoint_file = os.path.join(
                    checkpoint_path, f"checkpoint_{self.rank_info.global_rank}.pt"
                )
                self.assertTrue(os.path.exists(checkpoint_file))

                loaded = torch.load(checkpoint_file)
                self.assertEqual(loaded["epoch"], i)

        finally:
            checkpointer.close()

    def test_async_error_handling(self):
        """Test error handling in async operations."""
        # Create checkpointer with mocked components to simulate errors
        mock_stager = Mock()
        mock_process = Mock()
        mock_reader = Mock()

        # Mock staging to return a completed future
        mock_staging_future = Future()
        mock_staging_future.set_result({"staged": "data"})
        mock_stager.stage.return_value = mock_staging_future

        # Mock process write to raise an error
        mock_write_future = Future()
        mock_write_future.set_exception(RuntimeError("Write failed"))
        mock_process.write.return_value = mock_write_future

        checkpointer = AsyncCheckpointer(
            checkpoint_stager=mock_stager,
            checkpoint_process=mock_process,
            reader=mock_reader,
        )

        try:
            # This should not raise immediately
            stage_future, write_future = checkpointer.save("/tmp/test", self.state_dict)

            # But waiting for the write future should raise the error
            with self.assertRaises(RuntimeError) as cm:
                write_future.result()

            self.assertIn("Write failed", str(cm.exception))

        finally:
            checkpointer.close()

    def test_async_future_results(self):
        """Test the results returned by async futures."""
        checkpointer = self._create_async_checkpointer()
        checkpoint_path = os.path.join(self.temp_dir, "checkpoint_results")

        try:
            # Save checkpoint
            stage_future, write_future = checkpointer.save(
                checkpoint_path, self.state_dict
            )

            # Both futures should complete successfully
            stage_result = stage_future.result()
            write_result = write_future.result()

            # Stage result is wrapped by wrap_future() so it returns None on success
            # This is intentional - the stage_future indicates completion, not data access
            self.assertIsNone(stage_result)

            # Write result should be None (success indicator)
            self.assertIsNone(write_result)

        finally:
            checkpointer.close()


if __name__ == "__main__":
    run_tests()
